/***********************************************************************
 * Pool Controller V1
 * Ken Rumer
 * May 11, 2014
 *
 * Use this code to control pump, actuators and heater for a pool and spa.
 *   Performs actions sent to it over serial (usb).  Will be using PHPSerial
 *   library from the Raspberry Pi to send the messages.  The organization
 *   of the string is:
 *   0,1  command type (flow, pump, heat, lights, etc.)
 *     01 - flow
 *     02 - pump
 *     03 - heat
 *     04 - clean
 *     05 - water lights
 *     06 - yard lights
 *     07 - thermistors
 *   2,3  action number
 *     flow:        00,01          : pool,spa
 *     pump:        00,01,02,03,04 : off,1000,1750,2500,3250 RPM
 *     heat:        00,01,02       : off,furnace,solar
 *     clean:       00,01          : skimmer,cleaner
 *     waterlights: 00,01,02,03    : off,pool,spa,both
 *     yardlights:  00,01,02,03    : off,walkway,trees,all
 *     thermistors: 00,01,02       : air,flow,solar
 *
 * 2 timers are used, if a user hits a button, wait buttonWait to approve
 *   the button press, if nothing has been pressed for inactive seconds, 
 *   shut off the LCD
 * Grab Serial input, button input, update timers and run commands every
 *   loop
 *
 ************************************************************************/
#include <SimpleTimer.h>
#include <MuxShield.h>
#include <math.h>
#include <MyButton.h>
#include <MyRelay.h>
#include <MyThermistor.h>
#include <MyControl.h>
#include <SoftwareSerial.h>
#include <SerLCD.h>
#include <Flow.h>
#include <Pump.h>
#include <Heat.h>
#include <Clean.h>
#include <WaterLight.h>
#include <YardLight.h>
#include <Temperature.h>

// Debug options are 0 for none 1 for all
const int debugLevel = 1;

// Digital Out Pins, row 0
const int suctionPin = 0;       // Relay 0, LOW:from pool, HIGH:from spa
const int outputPin = 1;        // Relay 1, LOW:pool, HIGH:spa
const int solarPin = 2;         // Relay 2, LOW:use solar heat, HIGH:no solar
const int cleanPin = 3;         // Relay 3, LOW:robot, HIGH:skimmer
const int pumpR1Pin = 4;        // Relay 4, LOW:Pump INP1 on, HIGH:Pump INP1 off  --- See pump table
const int pumpR2Pin = 5;        // Relay 5, LOW:Pump INP2 on, HIGH:Pump INP2 off  --- See pump table
const int pumpR3Pin = 6;        // Relay 6, LOW:Pump INP3 on, HIGH:Pump INP3 off  --- See pump table
const int pumpR4Pin = 7;        // Relay 7, LOW:Pump off, HIGH:Pump on --- See pump table
const int furnacePin1 = 8;       // Relay 8, LOW:Heater on, HIGH:Heater off
const int furnacePin2 = 9;       // Relay 8, LOW:Heater on, HIGH:Heater off
const int poolLightPin = 10;     // Relay 9, LOW:Pool light on, HIGH:Pool Light off
const int spaLightPin = 11;     // Relay 10, LOW:Spa light on, HIGH:Spa Light off
const int walkwayLightPin = 12; // Relay 12, LOW:walkway light on, HIGH:off
const int treeLightPin = 13;    // Relay 13, LOW:tree lights on, HIGH:off
const int unused1Pin = 14;      // Relay 14, unused
const int unused2Pin = 15;      // Relay 15, unused

// Digital In Pins row 1
const int flowButtonPin = 9;
const int pumpButtonPin = 10;
const int heatButtonPin = 11;
const int cleanButtonPin = 12;
const int waterLightButtonPin = 13;
const int yardLightButtonPin = 14;
const int temperatureButtonPin = 15;

// Analog In Pins row 2
const int airTempPin = 0;
const int flowTempPin = 1;
const int solarTempPin = 2;

// Control values
const int flowType = 1;
const int pumpType = 2;
const int heatType = 3;
const int cleanType = 4;
const int waterLightType = 5;
const int yardLightType = 6;
const int temperatureType = 7;

// Timeouts
const int inactiveTimeout = 20000; //20 seconds
const int inputTimeout = 2000;  //2 seconds

// Global timer controls
int inactiveTimer;
int inputTimer;

// Serial
const int serialBaudRate = 9600;

// LCd
const int lcdPin = 3;

// Global variable in case something happened
int userInput;

// Global Objects
SerLCD lcd(lcdPin);
MuxShield *muxShield;
Flow *flow;
Pump *pump;
Heat *heat;
Clean *clean;
WaterLight *waterLight;
YardLight *yardLight;
Temperature *temperature;
SimpleTimer timer;

//-------------------------------------------------------------------------------------------
// setup
//   no args
//   Initialize everything...
//-------------------------------------------------------------------------------------------
void setup() {
  userInput = false;
  Serial.begin(serialBaudRate);
  lcd.setBrightness(1);
  lcd.noDisplay();
  muxShield = new MuxShield();
  muxShield->setMode(1, DIGITAL_OUT);
  muxShield->setMode(2, DIGITAL_IN_PULLUP);
  muxShield->setMode(3, ANALOG_IN);
  muxShield->digitalWriteMS(1, unused1Pin, HIGH);
  muxShield->digitalWriteMS(1, unused2Pin, HIGH);
  flow = new Flow(suctionPin, outputPin, flowButtonPin, flowType, muxShield);
  pump = new Pump(pumpR1Pin, pumpR2Pin, pumpR3Pin, pumpR4Pin, pumpButtonPin, pumpType, muxShield);
  heat = new Heat(furnacePin1, furnacePin2, solarPin, heatButtonPin, heatType, muxShield);
  clean = new Clean(cleanPin, cleanButtonPin, cleanType, muxShield);
  waterLight = new WaterLight(poolLightPin, spaLightPin, waterLightButtonPin, waterLightType, muxShield);
  yardLight = new YardLight(treeLightPin, walkwayLightPin, yardLightButtonPin, yardLightType, muxShield);
  temperature = new Temperature(airTempPin, flowTempPin, solarTempPin, temperatureButtonPin, temperatureType, muxShield);
  inputTimer = timer.setTimeout(inputTimeout, controlDevices);
  timer.disable(inputTimer);
  inactiveTimer = timer.setTimeout(inactiveTimeout, deactivateDisplay);
  timer.disable(inactiveTimer);
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
// loop
//   no args
//   Main loop runs forever
//-------------------------------------------------------------------------------------------
void loop()
{
  timer.run();
  readSerialBus();
  getButtonStates();
  processInputs();
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
// readSerialBus
//   no args
//   If there are at least 4 bytes ready to be read, read them
//     determine the controlType and inform that device it is about to do something
//-------------------------------------------------------------------------------------------
void readSerialBus() {
  if (Serial.available() >= 4) {
    char inputBuffer[4];
    int controlType = 0;
    int controlCode = 0;
    memset(inputBuffer, '\0', sizeof(inputBuffer));
    Serial.readBytes(inputBuffer, sizeof(inputBuffer));
    controlType = getControlType(inputBuffer);
    controlCode = getControlCode(inputBuffer);
    if (controlType == flow->getControlType()) {
      flow->setProcessing(controlCode);
      userInput = true;
    } else if (controlType == pump->getControlType()) {
      pump->setProcessing(controlCode);
      userInput = true;
    } else if (controlType == heat->getControlType()) {
      heat->setProcessing(controlCode);
      userInput = true;
    } else if (controlType == clean->getControlType()) {
      clean->setProcessing(controlCode);
      userInput = true;
    } else if (controlType == waterLight->getControlType()) {
      waterLight->setProcessing(controlCode);
      userInput = true;
    } else if (controlType == yardLight->getControlType()) {
      yardLight->setProcessing(controlCode);
      userInput = true;
    } else if (controlType == temperature->getControlType()) {
      temperature->setProcessing(controlCode);
      userInput = true;
    }
  }
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
// getButtonStates
//   no args
//   If a button has been pushed, this is done in order and only 1 button will be
//     read at a time.  Determine the controlType and inform that device it is
//     about to do something
//-------------------------------------------------------------------------------------------
void getButtonStates() {
  if (flow->getButtonState() == LOW) {
    flow->setProcessing(flow->nextControlCode());
    userInput = true;
  } else if (pump->getButtonState() == LOW) {
    pump->setProcessing(pump->nextControlCode());
    userInput = true;
  } else if (heat->getButtonState() == LOW) {
    heat->setProcessing(heat->nextControlCode());
    userInput = true;
  } else if (clean->getButtonState() == LOW) {
    clean->setProcessing(clean->nextControlCode());
    userInput = true;
  } else if (waterLight->getButtonState() == LOW) {
    waterLight->setProcessing(waterLight->nextControlCode());
    userInput = true;
  } else if (yardLight->getButtonState() == LOW) {
    yardLight->setProcessing(yardLight->nextControlCode());
    userInput = true;
  } else if (temperature->getButtonState() == LOW) {
    temperature->setProcessing(temperature->nextControlCode());
    userInput = true;
  }
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
// processInputs
//   no args
//   if a device is about to do something, enable the LCD and set the text to the
//     device that is changing and the action it is going to take
//-------------------------------------------------------------------------------------------
void processInputs() {
  if (userInput) {
    userInput = false;
    if (flow->getProcessing()) {
      activateDisplay(flow->getHeader(), flow->getControlText());
    } else if (pump->getProcessing()) {
      activateDisplay(pump->getHeader(), pump->getControlText());
    } else if (heat->getProcessing()) {
      activateDisplay(heat->getHeader(), heat->getControlText());
    } else if (clean->getProcessing()) {
      activateDisplay(clean->getHeader(), clean->getControlText());
    } else if (waterLight->getProcessing()) {
      activateDisplay(waterLight->getHeader(), waterLight->getControlText());
    } else if (yardLight->getProcessing()) {
      activateDisplay(yardLight->getHeader(), yardLight->getControlText());
    } else if (temperature->getProcessing()) {
      activateDisplay(temperature->getHeader(), temperature->getControlText());
    }
  }
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
// activateDisplay
//   line1 - what to print on line 1 of LCD
//   line2 - what to print on line 2 of LCD
//   Starts the inputTimer which will fire after the user hasn't pressed a button or
//     written to serial for inputTimeout seconds, meaning the user is done with input.
//     Starts the inactiveTimer which will fire when the user hasn't pressed a button
//     or written to serial for inactiveTimeout seconds, meaning the user is inactive.
//     Enables the LCD, and prints the text from the arguments on the screen
//-------------------------------------------------------------------------------------------
void activateDisplay(char *line1, char *line2) {
  timer.enable(inputTimer);
  timer.restartTimer(inputTimer);
  timer.enable(inactiveTimer);
  timer.restartTimer(inactiveTimer);
  lcd.setBrightness(30);
  lcd.display();
  lcdPrint(line1, line2);
  debugln("Reset inactive and input timer, turned display on");
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
// controlDevices
//   no args
//   This runs after inputTimeout has fired.  This means the user hasn't pressed a button
//     or sent a controlType and controlCode over serial for inputTimeout seconds.  This
//     function tells the devices it is time to do some work, and disables inputTimer for
//     until next input.
//-------------------------------------------------------------------------------------------
void controlDevices() {
  inputTimer = timer.setTimeout(inputTimeout, controlDevices);
  timer.disable(inputTimer);
  if (flow->getProcessing()) {
    Serial.print("@@");
    Serial.print(flow->controlDevice());
    Serial.println("@@");
  else if (pump->getProcessing()) {
    Serial.print("@@");
    Serial.print(pump->controlDevice());
    Serial.println("@@");
  else if (heat->getProcessing()) {
    Serial.print("@@");
    Serial.print(heat->controlDevice());
    Serial.println("@@");
  else if (clean->getProcessing()) {
    Serial.print("@@");
    Serial.print(clean->controlDevice());
    Serial.println("@@");
  else if (waterLight->getProcessing()) {
    Serial.print("@@");
    Serial.print(waterLight->controlDevice());
    Serial.println("@@");
  else if (yardLight->getProcessing()) {
    Serial.print("@@");
    Serial.print(yardLight->controlDevice());
    Serial.println("@@");
  else if (temperature->getProcessing()) {
    Serial.print("@@");
    Serial.print(temperature->controlDevice());
    Serial.println("@@");
  }
  debugln("Should have done something...");
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
// deactivateDisplay
//   no args
//   Disables inactiveTimer because the user has been inactive for inactiveTimeout.  Turns
//     LCD off.
//-------------------------------------------------------------------------------------------
void deactivateDisplay() {
  inactiveTimer = timer.setTimeout(inactiveTimeout, deactivateDisplay);
  timer.disable(inactiveTimer);
  lcd.setBrightness(1);
  lcd.noDisplay();
  debugln("User has been inactive, turned display off");
}
//-------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
// helper functions
//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
int getControlType(char *text) {
  char buffer[3];
  buffer[0] = text[0];
  buffer[1] = text[1];
  buffer[2] = '\0';
  return atoi(buffer);
}
int getControlCode(char *text) {
  char buffer[3];
  buffer[0] = text[2];
  buffer[1] = text[3];
  buffer[2] = '\0';
  return atoi(buffer);
}
void debug(char *text) {
  if (debugLevel > 0) {
    Serial.print(text);
  }
}
void debug(int num) {
  if (debugLevel > 0) {
    Serial.print(num);
  }
}
void debugln(char *text) {
  if (debugLevel > 0) {
    Serial.println(text);
  }
}
void debugln(int num) {
  if (debugLevel > 0) {
    Serial.println(num);
  }
}
void lcdPrint(char *line1, char *line2) {
  lcd.clear();
  debugln("Setting LCD text");
  if (line1) {
    lcd.selectLine(1);
    lcd.print(line1);
    Serial.println(line1);
  }
  if (line2) {
    lcd.selectLine(2);
    lcd.print(line2);
    Serial.println(line2);
  }
  delete line1;
  delete line2;
}
